/**
 *	Evolutionary Algorithms
 *
 *  Bsiepielprogramm zum Beitrag: 
 *  "Neuheit in evolutionären Algorithmen" von Eckhart Arnold,
 *  in: "Philosophie des Neuen" (hrsg. von Peter Seele),
 *  Wissenschaftliche Buchgesellschaft Darmstadt, 2008.
 *
 *	Copyright (c) 2008,  Eckhart Arnold
 *
 * Dieses Programm ist freie Software. Sie können es unter den
 * Bedingungen der GNU General Public License, wie von der Free
 * Software Foundation veröffentlicht, weitergeben und/oder
 * modifizieren, entweder gemäß Version 3 der Lizenz oder (nach
 * Ihrer Option) jeder späteren Version.
 *
 * Die Veröffentlichung dieses Programms erfolgt in der Hoffnung,
 * daß es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE
 * GARANTIE, sogar ohne die implizite Garantie der MARKTREIFE oder
 * der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Details finden
 * Sie in der GNU General Public License.
 *
 * Sie sollten ein Exemplar der GNU General Public License zusammen
 * mit diesem Programm erhalten haben. Falls nicht, siehe
 * <http://www.gnu.org/licenses/>.
 */

package evolution.examples;

import java.util.HashMap;
import java.util.Set;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.SWT;

/**
 * @author eckhart
 *
 */
public final class EvolutionaryAlgorithms {
	enum STATE { WAITING, RUNNING, FINISHED };
	//private static final String STEP_NAME = "Step: ";

	private static int BATCH_MODE_SIMULATION_DEPTH = 5;
	
	private static HashMap<String, ISimulation> exampleDictionary =
				new HashMap<String, ISimulation>();
	private static ISimulation	example = null;

	private static HashMap<String, TreeItem> nodes =
			    new HashMap<String, TreeItem>();
	private static Tree	      exampleTree;
	private static TreeItem   selectedItem;
	private static Button     startButton, stopButton, stepButton, resetButton;
	// private static Label   stepLabel;
	private static Text	      helpText, logText;

	private static Display    display;
	private static Shell      shell;
	private static MessageBox queryDialog;
	private static GC	      outputGC, monitorGC;
	private static Monitor    monitor = null;
	
	private static volatile SimulationThread simulation = null;
	private static volatile boolean inhibitOutput = false;
	private static volatile boolean redrawPending = false;
	private static volatile long stepCounter, stepDisplay;
	private static volatile int logIndex=0;
	private static volatile long displayTimeStamp;

	private static class SimulationThread extends Thread {
		private int steps;

		public SimulationThread(int steps) {
			this.steps = steps;
		}

		public void run() {
			Thread thisThread = Thread.currentThread();
			Runnable redrawTask = new Runnable() {
				public void run() { if (!shell.isDisposed()) displayState(); }
			};
			while (simulation == thisThread && !example.finished()
					&& steps != 0) {
				example.step();
				for (String m : example.queryMonitors())
					monitor.addValue(m, example.readMonitor(m));
				stepCounter++;
				if (!redrawPending && System.currentTimeMillis() - displayTimeStamp > 500) {
					redrawPending = true;
					display.asyncExec(redrawTask);
				}
				steps--;
			}
			simulation = null;
			display.asyncExec(redrawTask);
			if (example.finished()) {
				display.asyncExec(new Runnable() {
					public void run() {
						if (!shell.isDisposed()) {
							startButton.setEnabled(false);
							stopButton.setEnabled(false);
							stepButton.setEnabled(false);
							resetButton.setEnabled(true);
						}
					}
				});
			} else {
				display.asyncExec(new Runnable() {
					public void run() {
						if (!shell.isDisposed()) {
							startButton.setEnabled(true);
							stepButton.setEnabled(true);
							resetButton.setEnabled(true);
						}
					}
				});
			}
		}

		public void halt() {
			simulation = null;
			try {	join();  }
			catch (InterruptedException e) { };
		}
	}

	private static void addExample(ISimulation example,
									   String name, String variant) {
		TreeItem node, item;
		if (nodes.containsKey(name)) node = nodes.get(name);
		else {
			node = new TreeItem(exampleTree, 0);
			node.setText(name);
			node.setData(name);
			nodes.put(name, node);
		}
		item = new TreeItem(node, 0);
		item.setText(variant);
		String key = name + "." + variant;
		item.setData(key);
		exampleDictionary.put(key, example);
	}

	private static STATE getState() {
		assert example != null;
		if (simulation == null)
			if (example.finished()) return STATE.FINISHED;
			else return STATE.WAITING;
		else return STATE.RUNNING;
	}

	private static void displayState() {
		if (inhibitOutput) return;
		if (stepCounter > stepDisplay && !outputGC.isDisposed()) {
			example.drawState(outputGC);
			stepDisplay = stepCounter;
			// stepLabel.setText(STEP_NAME+stepDisplay);
		}
		if (!monitorGC.isDisposed()) {
			monitor.update(monitorGC);
		}
		String s = example.accessLog().substring(logIndex);
		int l = s.length();
		if (l > 0 && !logText.isDisposed()) {
			logText.append(s);
			logIndex += l;
		}
		redrawPending = false;
		displayTimeStamp = System.currentTimeMillis();
	}

	private static void start() {
		assert getState() == STATE.WAITING;
		startButton.setEnabled(false);
		stepButton.setEnabled(false);
		resetButton.setEnabled(false);
		simulation = new SimulationThread(-1);
		simulation.start();
		stopButton.setEnabled(true);
	}

	private static void stop() {
		if (getState() == STATE.RUNNING) {
			stopButton.setEnabled(false);
			simulation.halt();
		}
	}

	private static void interrupt() {
		inhibitOutput = true;
		stop();
		example.terminate();
	}
	
	private static void step() {
		assert getState() == STATE.WAITING;
		startButton.setEnabled(false);
		stepButton.setEnabled(false);
		resetButton.setEnabled(false);
		simulation = new SimulationThread(1);
		simulation.start();		
	}

	private static void reset() {
		assert getState() != STATE.RUNNING;
		example.reset();
		inhibitOutput = false;					
		stepCounter = 0;
		stepDisplay = 0;
		logIndex = 0;
		logText.setText("");		
		displayState();
		example.drawState(outputGC);
		monitor.redraw(monitorGC);	
		startButton.setEnabled(true);
		stopButton.setEnabled(false);
		stepButton.setEnabled(true);
	}

	private static interface Painter { void draw(GC gc); }
	
	private static void saveGraphics(String filename, Painter painter) {
		Image image = new Image(display, new Rectangle(0, 0, 1600, 1200));
		GC imageGC = new GC(image);
		ImageLoader loader = new ImageLoader();
		loader.data = new ImageData[1];
		painter.draw(imageGC);
		loader.data[0] = image.getImageData();
		if (!filename.toLowerCase().endsWith(".bmp")) filename += ".bmp";
		loader.save(filename, SWT.IMAGE_BMP);
		image.dispose();
	}
	
	private static void saveReport(String directory, String suffix) throws IOException {
		if (selectedItem.isDisposed()) return;
		FileWriter fw;
		File dir = new File(directory, selectedItem.getParentItem().getText());
		dir = new File(dir, selectedItem.getText());
		dir = new File(dir, suffix);
		
		if (!dir.exists()) {
			dir.mkdirs();
		} else if (!dir.isDirectory()) throw new IOException("Cannot create directory " + 
		    directory + ", because a file with that name already exists.");
		
		System.out.println("Saving report: "+dir.getPath());
		
		saveGraphics((new File(dir, "output")).getPath(),
				     new Painter() {
						public void draw(GC gc) { example.drawState(gc); }
		             });
		saveGraphics((new File(dir, "monitor")).getPath(),
			     new Painter() {
					public void draw(GC gc) { monitor.redraw(gc); }
	             });
		monitor.redraw(monitorGC);
		
		fw = new FileWriter(new File(dir, "log.txt"));
		fw.write(example.accessLog().toString());
		fw.close();
		
		fw = new FileWriter(new File(dir, "description.txt"));
		fw.write(example.getDescription());
		fw.close();
	}
	
	private static void batchMode(String[] examples, String directory) throws IOException {
		for (String e : examples) {				
			boolean breakFlag = false;			
			for (TreeItem node: exampleTree.getItems()) {
				for (TreeItem item: node.getItems()) {
					if (item.getData() == e) {
						selectedItem = item;
						breakFlag = true;
						break;
					}
				}
				if (breakFlag) break;
			}
			if (!breakFlag) {
				System.out.println("Example "+e+" not found in example tree!");
				continue;
			}
			exampleTree.setSelection(selectedItem);			
			example = exampleDictionary.get(e);
			helpText.setText(example.getDescription());
			monitor = new Monitor(example.queryMonitors());				
			reset();
			inhibitOutput = true;
			int a = 0, b;
			for (int i = 0; i <= BATCH_MODE_SIMULATION_DEPTH; i++) {
				b = a; a = (int)Math.pow(10.0, i);
				simulation = new SimulationThread(a - b);
				simulation.start();
				while (simulation != null && !shell.isDisposed()) {
					while (display.readAndDispatch());
					try {
						Thread.sleep(50);
					} catch (InterruptedException ie) {} ;
				}
				if (shell.isDisposed()) { simulation.halt(); return; }
				saveReport(directory, "stage_"+(new Integer(i)).toString());
			}	
			example.terminate();
		}
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		display = new Display();
		
		shell = new Shell(display);
		shell.setText("Neuheit in Evolutionären Algorithmen - Beispielsimulation");

		SashForm sash = new SashForm(shell, SWT.HORIZONTAL);
		exampleTree = new Tree(sash, SWT.BORDER);
		Composite showPane = new Composite(sash, SWT.BORDER);
		helpText = new Text(sash, SWT.READ_ONLY|SWT.MULTI|SWT.WRAP|SWT.BORDER);
		Color yellow = new Color(display, 255, 255, 128);
		helpText.setBackground(yellow);
		helpText.setText(About.infoString);
		int[] weights = {1, 3, 1}; sash.setWeights(weights);

		TabFolder outputTabs = new TabFolder(showPane, SWT.TOP);
		
		TabItem canvasItem = new TabItem(outputTabs, 0);
		canvasItem.setText("Output");
		Canvas outputArea = new Canvas(outputTabs, SWT.BORDER);
		outputGC = new GC(outputArea);
		outputGC.setAdvanced(false);
		if (outputGC.getAdvanced()) outputGC.setAntialias(SWT.ON);
		outputGC.setLineWidth(0);
		canvasItem.setControl(outputArea);
		
		TabItem monitorItem = new TabItem(outputTabs, 0);
		monitorItem.setText("Monitor");
		Canvas monitorArea = new Canvas(outputTabs, SWT.BORDER);
		monitorGC = new GC(monitorArea);
		monitorGC.setAdvanced(false);
		if (monitorGC.getAdvanced()) monitorGC.setAntialias(SWT.ON);
		monitorGC.setLineWidth(2);
		monitorItem.setControl(monitorArea);
		
		TabItem textItem = new TabItem(outputTabs, 0);
		textItem.setText("Log");
		logText = new Text(outputTabs, SWT.READ_ONLY|SWT.MULTI|SWT.WRAP|SWT.V_SCROLL);
		textItem.setControl(logText);
		
		Composite buttonRow = new Composite(showPane, 0);
		startButton = new Button(buttonRow, SWT.PUSH);
		startButton.setText("Start->");  startButton.setEnabled(false);
		stopButton = new Button(buttonRow, SWT.PUSH);
		stopButton.setText("Stop.");  stopButton.setEnabled(false);
		stepButton = new Button(buttonRow, SWT.PUSH);
		stepButton.setText("Step *");  stepButton.setEnabled(false);
		resetButton = new Button(buttonRow, SWT.PUSH);
		resetButton.setText("Reset!");  resetButton.setEnabled(false);
		// stepLabel = new Label(buttonRow, SWT.BORDER);
		// stepLabel.setText("  "); // STEP_NAME + "0     ");
		RowLayout rl = new RowLayout(SWT.HORIZONTAL); rl.justify = true;
		buttonRow.setLayout(rl);
		outputTabs.setLayoutData(new GridData(GridData.FILL_BOTH));
		buttonRow.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		showPane.setLayout(new GridLayout(1, true));

		queryDialog = new MessageBox(shell, SWT.YES|SWT.NO|SWT.CANCEL);
		queryDialog.setText("Stop running simulation?");
		queryDialog.setMessage("Another simulation is still running!\n"+
							   "Do you want to cancel the running simulation\n"+
			               	   "and start another simulation instead?");			

//		addExample(new SimpleTest(1, 100), "Test", "Counter 1-100");
		// addExample(new SimpleTest(201, 400), "Test", "Counter 201-400");
		
		addExample(new TSP(TSPData.QUICK_TEST), "Simulation", "Kleiner Test");

		addExample(new TSP(TSPData.SEP_CLUSTERS), "Simulation", "Einfach, m. Häufungen");
     	addExample(new TSP(TSPData.SEP_RANDOM), "Simulation", "Einfach, zufällig");

		addExample(new TSP(TSPData.SU_CLUSTERS), "Simulation", "Restgenom, m. Häufungen");
     	addExample(new TSP(TSPData.SU_RANDOM), "Simulation", "Restgenom, zufällig");     	
     	
		addExample(new TSP(TSPData.BEU_CLUSTERS), "Simulation", "Boltzmann, m. Häufungen");
		addExample(new TSP(TSPData.BEU_RANDOM), "Simulation", "Boltzmann, zufällig");	

		addExample(new TSP(TSPData.FEW_MUTATIONS_C), "Simulation", "Wenig Mutation, m. Häufungen");
		addExample(new TSP(TSPData.FEW_MUTATIONS_R), "Simulation", "Wenig Mutation, zufällig");				
		
		addExample(new TSP(TSPData.REFINED_CLUSTERS), "Simulation", "Verfeinert, m. Häufungen");
		addExample(new TSP(TSPData.REFINED_RANDOM), "Simulation", "Verferinert, zufällig");			

		addExample(new About(), "Über dieses Programm", "Programm Information");
		
		Set<String> batch = null;
		String path = "";
		for (String s: args) { 
			if (s.toLowerCase().contains("-batch")) {
				batch = exampleDictionary.keySet();
				for (String e: batch.toArray(new String[batch.size()])) 
					if (e.toLowerCase().contains("test")) batch.remove(e);
			} else if (s.toLowerCase().contains("-test")) {
				batch = exampleDictionary.keySet();
				for (String e: batch.toArray(new String[batch.size()])) 
					if (!e.toLowerCase().contains("test")) batch.remove(e);
				BATCH_MODE_SIMULATION_DEPTH = 3;			
			} else {
				if (path == "") {
					path = s;
					System.out.println("Output path is " + path);
				} else System.out.println("Unreckognized option: "+s+" !");
			}
		}	
				
		outputArea.addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent e) {
				if (example != null) example.drawState(outputGC);
			}
		});
		
		monitorArea.addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent e) {
				if (monitor != null) monitor.redraw(monitorGC);  
			}
		});

		if (batch == null) {
			exampleTree.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					TreeItem item = (TreeItem) e.item;
					TreeItem parent = item.getParentItem();
					if (parent != null) {
						if (example != null && stepCounter > 0) {
							if (queryDialog.open() != SWT.YES) {
								exampleTree.setSelection(selectedItem);
								return;
							}
							interrupt();
						}
						selectedItem = item;
						example = exampleDictionary.get(item.getData());
						helpText.setText(example.getDescription());
						monitor = new Monitor(example.queryMonitors());				
						reset();
					}
				}
			});
	
			startButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					start();
				}
			});
			stopButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					stop();
				}
			});
			stepButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					step();
				}
			});
			resetButton.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					reset();
				}
			});
		}
		
		shell.setLayout(new FillLayout());
		shell.open();
		
		if (batch == null) {
			while (!shell.isDisposed())
				if (!display.readAndDispatch()) display.sleep();
			if (simulation != null) simulation.halt();
		} else {
			try {
				batchMode(batch.toArray(new String[batch.size()]), path);
			} catch (IOException e) {
				System.out.println("Execption "+e.toString()+" caught!");
			}
		}
		
		display.dispose();
		yellow.dispose();
	}

}
